module reflection_module
implicit none

  public :: make_data_container
  public :: make_reflections
  public :: write_out_hkl_weight_file   
  public :: write_out_reflection_into   ! for debugging
  public :: cal_weight_matrix
  
  private :: Gaussian, cal_gaussian_profiles

  integer, parameter, private :: db = kind(1.0)  
  !integer, parameter, private :: db = kind(1.d0)
  

  type reflections
    integer, dimension(:), allocatable :: h, k, l
    real(db), dimension(:), allocatable :: Iout  ! extracted intensity for this Bragg reflection 
    real(db), dimension(:), allocatable :: x_bragg  ! either two-theta or TOF Bragg value
    real(db), dimension(:), allocatable :: fwhm ! in same units as x_bragg
    real(db), dimension(:), allocatable :: lor  ! lorentz factor
    
    real(db), dimension(:), allocatable :: left_cutoff   ! In units of two-theta or tof. Any data point
                                      ! to the left of this cutoff are not considered
                                      ! part of the profile for this reflection. Used
                                      ! when calculating the profiles and included here
                                      ! since why not. 
    real(db), dimension(:), allocatable :: right_cutoff  ! Also in units of either two-theta or tof.
    
    integer, dimension(:), allocatable :: profile_start  ! stores the position in the pattern from where
                              ! the profile for this reflection begins.
                              ! Note profile_start = 1 refers to the 1st point in the pattern
    integer, dimension(:), allocatable :: profile_end    ! stores where the profile ends
    
    ! Allocate space for storing profile values for this reflection.
    ! Notice only the first profile_end - profile_start + 1 members of
    ! this array actually get used.  
    real(db), dimension(:,:), allocatable :: profile_vals 
    
    double precision, dimension(:,:), allocatable :: weight   ! weight matrix
  end type reflections
  
  

  type data_container
    real, dimension(:), allocatable :: x_obs  ! either observed TOF or two-theta
    real, dimension(:), allocatable :: y_obs  ! observed counts in the pattern
    real, dimension(:), allocatable :: var  ! estimated variances for the observed counts
  end type data_container
  


contains

  !
  ! return data structure
  !
  function make_data_container(npts,x_obs,y_obs) result(container)
    integer, intent(in) :: npts
    real, dimension(:), intent(in) :: x_obs, y_obs
    type(data_container) :: container
  
    real :: scale   ! used for getting rough estimate of y_obs's errors
  
    allocate( container%x_obs( npts ) )
    allocate( container%y_obs( npts ) )
    allocate( container%var( npts ) )
    
    ! make sure that the data are ordered according to increasing
    ! value of x_obs. This is assumed in a search algorithm used
    ! to determine the individual peak ranges elsewhere
  
  
    scale = 100.0 / maxval(y_obs(1:npts))  ! 10000 means smallest error is 1%  
   
    print *, "scale = ", scale
    
    if (x_obs(1) < x_obs(npts)) then
      container%x_obs = x_obs(1:npts)
      container%y_obs = scale * y_obs(1:npts)    
    else
      container%x_obs = x_obs(npts:1:-1)
      container%y_obs = scale * y_obs(npts:1:-1)
    end if

    ! try to fill in some error estimates. Assumed here that all
    ! the y_obs's are none-zero and positive. However, the y_obs's
    ! may have be scaled in some useless way. 
    ! The errors are calculated using the following rough approach,
    ! which consist of first scaling the y_obs's so that the max y_obs
    ! is is 10000, then assume counting statistics on these values. 
    
    ! scale = multiply the y_obs's by this value so that the max y-obs
    ! value is 10000
    

    
    container%var = container%y_obs  ! so when scale > 1 the error get bigger as expected
    
    
    if (minval(container%var) <= 0) then 
        write(*,*) "ERROR in function make_data_container"
        write(*,*) "negative or zero variance"
        stop
    end if
    
  end function make_data_container
  
  
  !
  ! return an array of reflections
  !
  function make_reflections(nrefs,h,k,l,Iout,x_bragg,fwhm,lor,min_x_obs,max_x_obs, &
      max_profile_points,number_of_off_diagonal_weight_elements) result(refs)
    integer, intent(in) :: nrefs
    integer, dimension(:) :: h, k, l
    real(db), dimension(:), intent(in) :: Iout, x_bragg, fwhm, lor
    real, intent(in) :: min_x_obs, max_x_obs
    integer :: max_profile_points, number_of_off_diagonal_weight_elements
    type(reflections) :: refs
  
    integer :: n_good_refs = 0
    integer :: i
  
    ! only accept a reflection which x_bragg is in the data range
    
    do i = 1, nrefs
      if (x_bragg(i) > min_x_obs .and. x_bragg(i) < max_x_obs) then
        n_good_refs = n_good_refs + 1
      end if 
    end do 
  
    allocate( refs%h( n_good_refs ) )
    allocate( refs%k( n_good_refs ) )
    allocate( refs%l( n_good_refs ) )
    allocate( refs%Iout( n_good_refs  ) )
    allocate( refs%x_bragg( n_good_refs  ) )
    allocate( refs%fwhm( n_good_refs  ) )
    allocate( refs%lor( n_good_refs  ) )
    
    allocate( refs%left_cutoff( n_good_refs  ) )
    allocate( refs%right_cutoff( n_good_refs  ) )
    allocate( refs%profile_start( n_good_refs  ) )
    allocate( refs%profile_end( n_good_refs  ) )
    allocate( refs%profile_vals( n_good_refs, max_profile_points ) )
    
    ! To save memory here only allocate space for the weight matrix elements that
    ! are going to be returned in the .hkl_weight file. Thus the allocated weight
    ! matrix element refs%weight(i,j) refers to the weight matrix element i, i+j-1
    allocate( refs%weight( n_good_refs, number_of_off_diagonal_weight_elements + 1) )
    
    
    ! Populate over the reflections in the data range. Note that the reflection are 
    ! forced to be ordered in order of increasing x_bragg value. This 
    ! is later used in the search algorithm for locating the starting 
    ! and finishing points of the peak profiles.
    
    n_good_refs = 0
    
    if (x_bragg(1) < x_bragg(nrefs)) then 
      do i = 1, nrefs
        if (x_bragg(i) > min_x_obs .and. x_bragg(i) < max_x_obs) then
          n_good_refs = n_good_refs + 1
          refs%h(n_good_refs) = h(i)
          refs%k(n_good_refs) = k(i)
          refs%l(n_good_refs) = l(i)
          refs%Iout(n_good_refs) = Iout(i)
          refs%x_bragg(n_good_refs) = x_bragg(i)
          refs%fwhm(n_good_refs) = fwhm(i)
          refs%lor(n_good_refs) = lor(i)
        end if 
      end do   
    else
      do i = nrefs, 1,-1
        if (x_bragg(i) > min_x_obs .and. x_bragg(i) < max_x_obs) then
          n_good_refs = n_good_refs + 1
          refs%h(n_good_refs) = h(i)
          refs%k(n_good_refs) = k(i)
          refs%l(n_good_refs) = l(i)
          refs%Iout(n_good_refs) = Iout(i)
          refs%x_bragg(n_good_refs) = x_bragg(i)
          refs%fwhm(n_good_refs) = fwhm(i)
          refs%lor(n_good_refs) = lor(i)
        end if 
      end do     
    end if
    
  end function make_reflections  
  

  !
  ! fills in the profile value fields of a reflection container
  !
  subroutine cal_gaussian_profiles(refs, data_in)
    type(reflections), intent(inout) :: refs
    type(data_container), intent(in) :: data_in

    integer :: num_refs  ! number of reflections
    integer :: num_data  ! number of data points
    integer :: i, j
    integer :: j_got_to_left_profile_side ! used in loop below 
    integer :: j_got_to_right_profile_side ! used in loop below
    
    double precision :: min_x_obs  ! min measured two-theta or tof value
    double precision :: max_x_obs
    
    num_refs = size(refs%h)
    num_data = size(data_in%x_obs)

    min_x_obs = minval(data_in%x_obs)
    max_x_obs = maxval(data_in%x_obs)
    

    ! calculate the left and right cutoff. used later below. 
        
    refs%left_cutoff = refs.x_bragg - 4.0 * refs.fwhm
    refs%right_cutoff = refs.x_bragg + 4.0 * refs.fwhm


    ! calculate the rest
    
    j_got_to_left_profile_side = 1
    j_got_to_right_profile_side = 1
    
    do i = 1, num_refs
    
      if (refs%right_cutoff(i) < min_x_obs .or. refs%left_cutoff(i) > max_x_obs) then
        ! because of the way the data and reflection info are read in
        ! using make_data_container() and make_reflections() this exception
        ! should never occur.
        write(*,*) "ERROR in subroutine cal_gaussian_profiles"
        write(*,*) "reflection outside data range"
        stop
      end if


      ! find first data point to the right of left cutoff

      do j = j_got_to_left_profile_side, num_data 
        if (data_in%x_obs(j) > refs%left_cutoff(i)) then
          j_got_to_left_profile_side = j
          refs%profile_start(i) = j
          exit
        end if 
      end do
      
      ! find first data point to the left of right cutoff

      do j = j_got_to_right_profile_side, num_data 
        if (data_in%x_obs(j) > refs%right_cutoff(i)) then
          j_got_to_right_profile_side = j-1   ! minus one because we what to 
                                              ! refer to the first point just 
                                              ! to the left of the right cutoff 
          refs%profile_end(i) = j-1
          exit
        end if 
      end do
      
      ! if j == num_data+1 this means that the right_cutoff is to the right of the
      ! last data point, therefore for this case assume that the profile ends at
      ! the last data point for this case
      
      if (j == num_data + 1) then
        j_got_to_right_profile_side = num_data
        refs%profile_end(i) = num_data
      end if
       
      
      ! calculate profile values from profile_start to profile_end
      
      do j = 1, refs%profile_end(i)-refs%profile_start(i) + 1
        refs%profile_vals(i,j) = gaussian( refs%x_bragg(i)-data_in%x_obs(refs%profile_start(i)-1+j), refs%fwhm(i) );     
      end do  
            
    end do

  end subroutine cal_gaussian_profiles
  
  
  !
  ! calculate weight matrix elements
  !
  subroutine cal_weight_matrix(refs, data_in)
    type(reflections), intent(inout) :: refs
    type(data_container), intent(in) :: data_in
  
    integer :: num_refs  ! number of reflections
    integer :: i, j, jj
    
    integer :: num_profile_val  ! used in loop below
    integer :: profile_start_i  
    integer :: profile_end_jj
    
    integer :: number_of_off_diagonal_weight_elements
    
    num_refs = size(refs%h) 
  
    number_of_off_diagonal_weight_elements = size(refs%weight,2) - 1
    
    
    ! before calculating weight matrix elements calculate profiles
    
    call cal_gaussian_profiles(refs, data_in)
  
  
    ! initialise weight matrix
    
    refs%weight = 0.0
      
  
    ! calculate diagonal weight matrix elements
      
    do i = 1, num_refs
      num_profile_val = refs%profile_end(i) - refs%profile_start(i) + 1
      
      refs%weight(i, 1) = sum(refs%profile_vals(i, 1:num_profile_val)**2 / &
        data_in%var(refs%profile_start(i):refs%profile_end(i)))
    end do


    ! Populate the off-diagonal weight matrix elements
  
    do i = 1, num_refs
      do j = 2, number_of_off_diagonal_weight_elements + 1
      
        jj = i + j - 1   ! this refer to reflection jj. Hence the purpose of
                         ! the code here is the calculate the weight matrix 
                         ! element W(i,jj) which we store as refs%weight(i,j)
        
        if (jj > num_refs) then
          exit
        end if
        
        
        ! if no overlap go to next element
        
        if ( refs%profile_start(jj) > refs%profile_end(i) ) then
          exit
        end if

        
        ! The line below calculate the 1st profile point for the i'th reflection,
        ! which is none-zero for both the i'th reflection and the jj'th reflection
        
        profile_start_i = refs%profile_start(jj) - refs%profile_start(i) + 1
        
        
        ! The line below calculate the last profile point for the jj'th reflection,
        ! which is none-zero for both the i'th reflection and the jj'th reflection
        
        profile_end_jj = refs%profile_end(i) - refs%profile_start(jj) + 1
        
        
        ! sum over the common none-zero region of reflections i and j to calculate the 
        ! off-diagonal weight matrix element between reflections i and j

        num_profile_val = refs%profile_end(i) - refs%profile_start(i) + 1
        
        refs%weight(i,j) = sum( refs%profile_vals(i,profile_start_i:num_profile_val) * &
          refs%profile_vals(jj,1:profile_end_jj) / data_in%var(refs%profile_start(jj):refs%profile_end(i)) )
      end do
    end do
      
  end subroutine cal_weight_matrix
  
  
  !
  ! Gaussian peak shape function. Copied from the CrysFML library.
  !
  ! Note FWHM = 2*sqrt(2*ln(2))*sigma = 2.3548*sigma. This means that
  ! 1/sqrt(2*pi*sigma^2) = 2*sqrt(ln(2)/pi)*FWHM = 0.9394.... * FWHM (hence ag below ok)
  ! 1/(2*sigma^2)=4*ln(2)/(FWHM^2) = 2.77.... /(FWHM^2) (hence bg ok below)
  !
  Pure Function Gaussian(X,H) Result (Gauss_Val)
      !---- Arguments ----!
      real,              intent(in) :: x
      real,              intent(in) :: H
      real                          :: gauss_val

      !--- Local variables ---!
      real :: ag,bg

      ag= 0.93943727869965133377234032841018/H
      bg= 2.7725887222397812376689284858327/(H*H)
      gauss_val = ag* exp(-bg*x*x)

      return
  End Function Gaussian  
    
    
  !
  ! for debugging
  !  
  subroutine write_out_reflection_into(filename, refs)  
    character(*), intent(in) :: filename
    type(reflections), intent(in) :: refs
    
    integer :: num_refs, i
    
    num_refs = size(refs%h)
    
    
    open(11,file=trim(filename),status='unknown')    
    
    do i = 1, num_refs       
	    write(11,1100) refs%h(i), refs%k(i), refs%l(i), refs%Iout(i), refs%x_bragg(i), refs%fwhm(i)
	  end do    

	1100 format(3i5,3f12.4)
    
    close(11)    
  end subroutine write_out_reflection_into 
 
    
  !
  ! Write to disk a .hkl_weight file 
  !
  subroutine write_out_hkl_weight_file(filename, refs, scale)
    character(*), intent(in) :: filename
    type(reflections), intent(in) :: refs
    double precision :: scale
    
    integer :: num_refs, i, j
    
    integer :: number_of_off_diagonal_weight_elements
    
    num_refs = size(refs%h) 
  
    number_of_off_diagonal_weight_elements = size(refs%weight,2) - 1
    
    
    open(11,file=trim(filename),status='unknown')
	
	    
    do i = 1, num_refs
      write(11,1200, advance='no') refs%h(i), refs%k(i), refs%l(i), refs%Iout(i)*scale,refs%lor(i)*sqrt(refs%weight(i,1)), i 
      
      do j = 1, number_of_off_diagonal_weight_elements - 1
        if (i+j <= num_refs) then
        
          ! Notice here refs%weight(i,1) refers to diagonal matrix element i, i
          ! Notice here refs%weight(i+j,1) refers to diagonal matrix element i+j, i+j
          ! Notice here refs%weight(i,j+1) refers to off-diagonal matrix element i, i+j
          ! 
          write(11,1300, advance='no') refs%weight(i,j+1)*100.0 / sqrt(refs%weight(i,1)*refs%weight(i+j,1))
        else
          write(11,1300, advance='no') 0
        end if
      end do
      
      ! and to get the bloody carrier return the last off-diagonal element is written out separately
      ! without the anvange='no' flag
      
      j = number_of_off_diagonal_weight_elements
      if (i+j <= num_refs) then
        write(11,1300) refs%weight(i,j+1)*100.0 / sqrt(refs%weight(i,1)*refs%weight(i+j,1))
      else
        write(11,1300) 0
      end if      
      
    end do
    
 1200 format(3i5,f14.5,f14.6,i6,x)
 1300 format(f6.1,x)  
    
  close(11)
  
  end subroutine write_out_hkl_weight_file 
    
    
end module reflection_module
